"""CSSStyleDeclaration implements DOM Level 2 CSS CSSStyleDeclaration.

see
    http://www.w3.org/TR/1998/REC-CSS2-19980512/syndata.html#parsing-errors

Unknown properties
------------------
User agents must ignore a declaration with an unknown property.
For example, if the style sheet is::

    H1 { color: red; rotation: 70minutes }

the user agent will treat this as if the style sheet had been::

    H1 { color: red }

Cssutils gives a warning about an unknown CSS2 Property "rotation".

Illegal values
--------------
User agents must ignore a declaration with an illegal value. For example::

    IMG { float: left }       /* correct CSS2 */
    IMG { float: left here }  /* "here" is not a value of 'float' */
    IMG { background: "red" } /* keywords cannot be quoted in CSS2 */
    IMG { border-width: 3 }   /* a unit must be specified for length values */

A CSS2 parser would honor the first rule and ignore the rest, as if the
style sheet had been::

    IMG { float: left }
    IMG { }
    IMG { }
    IMG { }

Cssutils again will issue warning about invalid CSS2 property values.

TODO:
    This interface is also used to provide a read-only access to the
    computed values of an element. See also the ViewCSS interface.

    - return computed values and not literal values
    - simplify unit pairs/triples/quadruples
      2px 2px 2px 2px -> 2px for border/padding...
    - normalize compound properties like:
      background: no-repeat left url()  #fff
      -> background: #fff url() no-repeat left      
"""
__all__ = ['CSSStyleDeclaration']
__docformat__ = 'restructuredtext'
__version__ = '0.9a6'

import xml.dom 

import cssutils
from property import _Property as Property


class CSSStyleDeclaration(cssutils.util.Base):
    """
    The CSSStyleDeclaration class represents a single CSS declaration
    block. This class may be used to determine the style properties
    currently set in a block or to set style properties explicitly
    within the block.
    
    While an implementation may not recognize all CSS properties within
    a CSS declaration block, it is expected to provide access to all 
    specified properties in the style sheet through the 
    CSSStyleDeclaration interface. 
    Furthermore, implementations that support a specific level of CSS
    should correctly handle CSS shorthand properties for that level. For
    a further discussion of shorthand properties, see the CSS2Properties
    interface.

    Properties
    ==========
    cssText: of type DOMString
        The parsable textual representation of the declaration block
        (excluding the surrounding curly braces). Setting this attribute
        will result in the parsing of the new value and resetting of the
        properties in the declaration block. It also allows the insertion
        of additional properties and their values into the block.
    length: of type unsigned long, readonly
        The number of properties that have been explicitly set in this
        declaration block. The range of valid indices is 0 to length-1
        inclusive.
    parentRule: of type CSSRule, readonly
        The CSS rule that contains this declaration block or None if this
        CSSStyleDeclaration is not attached to a CSSRule.
    seq: a list (cssutils)
        All parts of this style declaration including CSSComments

    Format
    ======
    [Property: Value;]* Property: Value?    
    """
    def __init__(self, parentRule=None, cssText=u'', readonly=False):
        """
        parentRule
            The CSS rule that contains this declaration block or
            None if this CSSStyleDeclaration is not attached to a CSSRule.
        readonly
            defaults to False        
        """
        self.seq = []
        self.parentRule = parentRule
        self.cssText = cssText
        self._readonly = readonly
        

    def _getCssText(self):
        """
        returns serialized property cssText
        """
        return cssutils.ser.do_css_CSSStyleDeclaration(self)

    def _setCssText(self, cssText):
        """
        Setting this attribute will result in the parsing of the new value
        and resetting of all the properties in the declaration block
        including the removal or addition of properties.

        DOMException on setting
        
        - NO_MODIFICATION_ALLOWED_ERR: (self)
          Raised if this declaration is readonly or a property is readonly.
        - SYNTAX_ERR: (self)
          Raised if the specified CSS string value has a syntax error and
          is unparsable.          
        """
        def ignoreuptopropend(i, tokens):
            "returns position of ending ;"
            ignoredtokens, endi = tokenizer.tokensupto(
                    tokens[i:], propertypriorityendonly=True)
            if ignoredtokens:
                ignored = ''.join([x.value for x in ignoredtokens])
                log.error(u'CSSStyleDeclaration: Ignored: "%s".' %
                          self._valuestr(ignoredtokens))
            return endi + 1

        self._checkReadonly()
        tokenizer = cssutils.tokenize.Tokenizer()
        tokens, ttypes, log = self._prepare(cssText)

        newseq = []
        i = 0
        while i < len(tokens):
            t = tokens[i]
            if ttypes.S == t.type: # ignore
                pass
            elif ttypes.COMMENT == t.type: # just add
                newseq.append(cssutils.css.CSSComment(t))            
            else:
                # name upto ":" (or ; -> error)
                nametokens, endi = tokenizer.tokensupto(
                    tokens[i:], propertynameendonly=True)
                i += endi

                shouldbecolon = nametokens.pop()
                if shouldbecolon.value == u':': # OK: exclude ending :
                    i += 1 
                elif shouldbecolon.value == u';': # ERROR: premature ;
                    log.error(
                        u'CSSStyleDeclaration: Incomplete Property starting here: %s.' %
                        self._valuestr(tokens[i-1:]))
                    i += 1 # exclude ending :
                    continue
                else: # ERROR: no :
                    log.error(
                        u'CSSStyleDeclaration: No Propertyname and/or ":" found: %s.' %
                              self._valuestr(tokens[i:]))
                    i += ignoreuptopropend(i, tokens)
                    continue

                nameident = None
                for x in nametokens:
                    if x.type == ttypes.IDENT:
                        nameident = x.value
                        break
                if not nameident: # ERROR: no name
                    log.error(
                        u'CSSStyleDeclaration: No Propertyname found: %s.'
                        % self._valuestr(tokens[i-1:]))
                    i += ignoreuptopropend(i, tokens)
                    continue
                
                # value upto ";" or "!important" or end
                valuetokens, endi = tokenizer.tokensupto(
                    tokens[i:], propertyvalueendonly=True)
                i += endi
                if valuetokens and \
                   valuetokens[-1].type == ttypes.SEMICOLON:
                    del valuetokens[-1] # exclude ending ;            
                    prioritytokens = None
                elif valuetokens and \
                     valuetokens[-1].type == ttypes.IMPORTANT_SYM:
                    del valuetokens[-1] # exclude !important

                    # priority upto ; or end                    
                    prioritytokens, endi = tokenizer.tokensupto(
                            tokens[i:], propertypriorityendonly=True)
                    i += endi
                    
                    if prioritytokens and prioritytokens[-1].type == \
                       ttypes.SEMICOLON:
                        del prioritytokens[-1] # exclude ending ;
                elif not valuetokens:
                    log.error(u'CSSStyleDeclaration: No property value: %s'
                              % self._valuestr(cssText))
                    i += ignoreuptopropend(i, tokens)
                    continue
                else:
                    prioritytokens = None

                self.setProperty(nametokens, valuetokens, prioritytokens,
                                 _seq=newseq, _nameident=nameident)
                    
            i += 1
            
        self.seq = newseq

    cssText = property(_getCssText, _setCssText,
        doc="(DOM) The parsable textual representation of the declaration\
        block excluding the surrounding curly braces.")


    def _getLength(self):
        return len([x for x in self.seq if isinstance(x, Property)])
    
    length = property(_getLength,
        doc="(DOM) the number of properties that have been explicitly set\
        in this declaration block. The range of valid indices is 0 to\
        length-1 inclusive.")


    def _getParentRule(self):
        return self._parentRule
        
    def _setParentRule(self, parentRule):
        self._parentRule = parentRule

    parentRule = property(_getParentRule, _setParentRule,
        doc="(DOM) The CSS rule that contains this declaration block or\
        None if this CSSStyleDeclaration is not attached to a CSSRule.")

    
    def getPropertyCSSValue(self, name):
        """
        (DOM)
        Used to retrieve the object representation of the value of a CSS
        property if it has been explicitly set within this declaration
        block. This method returns None if the property is a shorthand
        property. Shorthand property values can only be accessed and
        modified as strings, using the getPropertyValue and setProperty
        methods.

        name
            of the CSS property
        
        returns CSSValue, the value of the property if it has been
        explicitly set for this declaration block. Returns None if the
        property has not been set.
        """
        for p in self.seq:
            if isinstance(p, Property) and p.name == name.lower():
                return p.value


    def getPropertyValue(self, name):
        """
        (DOM)
        Used to retrieve the value of a CSS property if it has been
        explicitly set within this declaration block.
        
        name
            of the CSS property
        
        returns the value of the property if it has been explicitly set
        for this declaration block. Returns the empty string if the
        property has not been set. 
        """
        for p in self.seq:
            if isinstance(p, Property) and p.name == name.lower():
                return p.value #.cssText
        return u''


    def getPropertyPriority(self, name):
        """
        (DOM)
        Used to retrieve the priority of a CSS property (e.g. the
        "important" qualifier) if the property has been explicitly set in
        this declaration block.
        
        name
            of the CSS property
        
        returns a string representing the priority (e.g. "important") if
        one exists. The empty string if none exists.
        """
        for p in self.seq:
            if isinstance(p, Property) and p.name == name.lower():
                return p.priority
        return u''


    def item(self, index):
        """
        (DOM)
        Used to retrieve the properties that have been explicitly set in
        this declaration block. The order of the properties retrieved using
        this method does not have to be the order in which they were set.
        This method can be used to iterate over all properties in this
        declaration block.

        index
            of the property to retrieve, negative values behave like
            negative indexes on Python lists, so -1 is the last element
        
        returns the name of the property at this ordinal position. The
        empty string if no property exists at this position.
        """
        properties = [x for x in self.seq if isinstance(x, Property)]
        try:
            return properties[index].name
        except IndexError:
            return u''


    def removeProperty(self, name):
        """
        (DOM)
        Used to remove a CSS property if it has been explicitly set within
        this declaration block.
        
        name
            of the CSS property to remove
        
        returns the value of the property if it has been explicitly set for
        this declaration block. Returns the empty string if the property
        has not been set or the property name does not correspond to a
        known CSS property
        
        raises DOMException
        
        - NO_MODIFICATION_ALLOWED_ERR: (self)
          Raised if this declaration is readonly or the property is
          readonly.
        """
        self._checkReadonly()
        
        r = u''
        for i, p in enumerate(self.seq):
            if isinstance(p, Property) and p.name == name:
                r = p.value #.cssText
                del self.seq[i]
                break        
                    
        return r


    def setProperty(self, name, value, priority=None,
                    _seq=None, _nameident=None):
        """
        (DOM)
        Used to set a property value and priority within this declaration
        block.
        
        name
            of the CSS property to set
            (in W3C DOM the parameter is called "propertyName")
        value
            the new value of the property
        priority
            the optional priority of the property (e.g. "important")
        _seq, _nameident
            used be self.cssText only
            
        DOMException on setting
        
        - SYNTAX_ERR: (self)
          Raised if the specified value has a syntax error and is
          unparsable.
        - NO_MODIFICATION_ALLOWED_ERR: (self)
          Raised if this declaration is readonly or the property is
          readonly.
        """
        self._checkReadonly()

        if _seq is None: # seq to update (maybe newseq in ``cssText``)
            _seq = self.seq
        if _nameident is None: # name to check if set already
            _nameident = name
            
        for p in _seq:
            if isinstance(p, Property) and p.name == _nameident.lower():
                p.value = value
                p.priority = priority
                break
        else:
            p = Property(name, value, priority)
            if p.name and p.value:
                _seq.append(p)


if __name__ == '__main__':
    cssutils.css.cssstyledeclaration.Property = Property
    
    s = cssutils.parseString('''a {
        c\olor: green
    }''')#color:green
    print s.cssText

