"""Selector is a single Selector of a CSSStyleRule SelectorList.

TODO
    - .contains(selector)
    - .isSubselector(selector)

    - ??? CSS2 gives a special meaning to the comma (,) in selectors.
        However, since it is not known if the comma may acquire other
        meanings in future versions of CSS, the whole statement should be
        ignored if there is an error anywhere in the selector, even though
        the rest of the selector may look reasonable in CSS2.

        Illegal example(s):

        For example, since the "&" is not a valid token in a CSS2 selector,
        a CSS2 user agent must ignore the whole second line, and not set
        the color of H3 to red:
"""
__all__ = ['Selector']
__docformat__ = 'restructuredtext'
__version__ = '0.9a5'

import xml.dom

import cssutils


class Selector(cssutils.util.Base):
    """
    (cssutils) a single selector in a SelectorList of a CSSStyleRule

    Properties
    ==========
    selectorText
        textual representation of this Selector
    seq
        sequence of Selector parts including comments
    
    Format
    ======
    combinator
      : PLUS S* | GREATER S* | S+ ;
    selector
      : simple_selector [ combinator simple_selector ]* ;
    simple_selector
      : element_name [ HASH | class | attrib | pseudo ]*
      | [ HASH | class | attrib | pseudo ]+ ;
    class 
      : '.' IDENT ;
    element_name
      : IDENT | '*' ;
    attrib
      : '[' S* IDENT S* [ [ '=' | INCLUDES | DASHMATCH ] S*
      [ IDENT | STRING ] S* ]? ']'
      ;
    pseudo
      : ':' [ IDENT | FUNCTION S* IDENT? S* ')' ] ;

    plus some from http://www.w3.org/TR/css3-selectors/      
    """
    def __init__(self, selectorText=None, readonly=False):
        """
        selectorText
            initial value of this selector
        readonly
            default to False
        """
        self.seq = []
        if selectorText:
            self.selectorText = selectorText
        self._readonly = readonly


    def _getSelectorText(self):
        """ 
        returns serialized format 
        """
        return cssutils.ser.do_css_Selector(self)

    def _setSelectorText(self, selectorText):
        """
        DOMException on setting

        - INVALID_MODIFICATION_ERR: (self)
        - INVALID_MODIFICATION_ERR: (self)
          Raised if the specified CSS string value represents a different
          type than the current one, e.g. a SelectorList.              
        - NO_MODIFICATION_ALLOWED_ERR: (self)
          Raised if this rule is readonly.
        - SYNTAX_ERR: (self)
          Raised if the specified CSS string value has a syntax error
          and is unparsable.     
        """
        def preprocess(tokens):
            """
            remove S at start and end,
            if around a combinator (ignoring comments)
            or in att []
            """
            # remove starting S
            if tokens[0].type == ttypes.S:
                del tokens[0]
            COMBINATORS_NO_S = u'+>~'
            r = []
            attlevel = 0
            i = 0
            while i < len(tokens):
                t = tokens[i]
                if ttypes.S == t.type and attlevel > 0:
                    # remove all WS in atts 1st level
                    i += 1
                    continue

                elif t.value in COMBINATORS_NO_S and attlevel == 0:
                    # remove preceding S from found except token!=Comment
                    before = len(r) 
                    while before >= 0:
                        before -= 1
                        if r[before].type == ttypes.S:
                            del r[before]
                        elif r[before].type != ttypes.COMMENT:
                            break
                    # remove following S from tokens except token!=Comment
                    after = i+1 
                    while after < len(tokens):
                        if tokens[after].type == ttypes.S:
                            del tokens[after]
                            after -= 1 # BECAUSE DELETED ONE!
                        elif tokens[after].type != ttypes.COMMENT:
                            break
                        after += 1                        
                elif t.value == u'[':
                    attlevel += 1
                elif t.value == u']':
                    attlevel -= 1
                r.append(t)
                i += 1

            # remove ending S
            if r and r[-1].type == ttypes.S:
                del r[-1]
            return r
            
        def getAttrib(atttokens):
            """
            atttokens
                a sequence of tokens without any S which are preprocessed

            returns seq of selector attribute

            Format
            ======
            attrib
              : '[' S* IDENT S* [ [ '=' | INCLUDES | DASHMATCH ] S*
                [ IDENT | STRING ] S* ]? ']' ;

            CSS3 additionally:
                ^= PREFIXMATCH 
                $= SUFFIXMATCH
                *= SUBSTRINGMATCH 
            """
            attseq = []
            expected = "[" # att, comb or ], value
            i = 0
            while i < len(atttokens):
                t = atttokens[i]
                if ttypes.COMMENT == t.type: # just add
                    attseq.append(cssutils.css.CSSComment(t))

                elif expected in ('comb or ]', ']') and t.value == u']':
                    attseq.append(t.value)
                    assert i == len(atttokens) - 1 # should really be end
                    break

                elif expected == '[' and t.value == u'[':
                    attseq.append(t.value)
                    expected = 'att'
                elif expected == '[':
                    log.error(u'Selector: No Attrib start "[" found.', t)
                    return

                elif expected == 'att' and t.type == ttypes.IDENT:
                    attseq.append(t.value)
                    expected = 'comb or ]'
                elif expected == 'att':
                    log.error(u'Selector: No Attrib found.', t)
                    return
                
                # CSS3 selectors added
                elif expected == 'comb or ]' and t.value in (
                     u'=', u'~=', u'|=', u'^=', u'$=', u'*='):
                    attseq.append(t.value)
                    expected = 'val'
                elif expected == 'comb or ]':
                    log.error(u'Selector: No Attrib combinator found.', t)
                    return

                elif expected == 'val' and t.type in (
                     ttypes.IDENT, ttypes.STRING):
                    attseq.append(t.value)
                    expected = ']'
                elif expected == 'val':
                    log.error(u'Selector: No Attrib value found.', t)
                    return

                else:
                    log.error(u'Selector: Unknown Attrib syntax.', t)
                    return

                i += 1

            else: # in case of ] not found
                log.error(u'Selector: Incomplete Attrib.', t)
                return
            
            return attseq


        def getFunc(functokens):
            """
            functokens
                a sequence of tokens

            returns seq of selector functional_pseudo

            Format
            ======
            functional_pseudo
              : FUNCTION S* expression ')'
              ;
            """
            funcseq = []
            expected = ")" # expr or )
            i = 0
            while i < len(functokens):
                t = functokens[i]
                if ttypes.COMMENT == t.type: # just add
                    funcseq.append(cssutils.css.CSSComment(t))

                elif t.value == u')':
                    funcseq.append(t.value)
                    expected = None
                    assert i == len(functokens) - 1 # should really be end
                    break

                else:
                    funcseq.append(t.value)

                i += 1

            if expected:
                log.error(
                    u'Selector: Incomplete functional pseudo "%s".' %
                    self._valuestr(functokens))
                return
            
            return funcseq


        self._checkReadonly()
        tokenizer = cssutils.tokenize.Tokenizer()
        tokens, ttypes, log = self._prepare(selectorText)

        if tokens:
            COMBINATORS = u' >+~'
            tokens = preprocess(tokens)
            
            newseq = []
            found1 = False
            i = 0
            expected = 'simple_selector' # simple_selector or combinator
            while i < len(tokens):
                t = tokens[i]
                if ttypes.COMMENT == t.type and expected not in (
                     'classname', 'pseudo'): # just add
                    newseq.append(cssutils.css.CSSComment(t))

                elif ttypes.COMMA == t.type:
                    log.error(u'Selector: Possible SelectorList found.', t,
                              error=xml.dom.InvalidModificationErr)
                    return

                elif expected.startswith('simple_') and \
                     t.type == ttypes.CLASS:
                    newseq.append(t.value)
                    found1 = True
                    expected = 'classname'
                elif expected == 'classname' and t.type == ttypes.IDENT:
                    newseq.append(t.value)
                    expected = 'simple_selector or combinator'
                    
                elif expected.startswith('simple_') and \
                     t.value == u':' or t.value == u'::':
                    newseq.append(t.value)
                    found1 = True
                    expected = 'pseudo'
                elif expected == 'pseudo' and t.type == ttypes.IDENT:
                    newseq.append(t.value)
                    expected = 'simple_selector or combinator'
                elif expected == 'pseudo' and t.type == ttypes.FUNCTION:
                    # function like lang(...)
                    functokens, endi = tokenizer.tokensupto(
                        tokens[i:], selectorattendonly=True)
                    i += endi
                    funcseq = getFunc(functokens)
                    if not funcseq:
                        log.error(
                            u'Selector: Invalid functional pseudo.', t)
                        return
                    newseq.extend(funcseq)
                    expected = 'simple_selector or combinator'

                elif expected == 'simple_selector' and (
                     t.type in (ttypes.IDENT, ttypes.HASH) or \
                     t.value in u'*'):
                    newseq.append(t.value)
                    found1 = True
                    expected = 'simple_selector or combinator'

                elif expected == 'simple_selector or combinator' and (
                     t.type in (ttypes.HASH) or t.value in u'*'):
                    newseq.append(t.value)
                    found1 = True
                    expected = 'simple_selector or combinator'

                elif expected.startswith('simple_') and t.value == u'[':
                    atttokens, endi = tokenizer.tokensupto(
                        tokens[i:], selectorattendonly=True)
                    i += endi
                    attseq = getAttrib(atttokens)
                    if not attseq:
                        log.warning(
                            u'Selector: Empty attribute selector.', t)
                        return
                    newseq.extend(attseq)
                    found1 = True
                    expected = 'simple_selector or combinator'

                # CSS3 combinator ~
                elif expected == 'simple_selector or combinator' and \
                     t.value in COMBINATORS:
                    newseq.append(t.value)
                    expected = 'simple_selector' # or [
                elif expected == 'simple_selector or combinator':
                    log.error(u'Selector: No combinator found.', t)
                    return

                else:
                    log.error(u'Selector: Expected simple selector.', t)
                    return

                i += 1
                
            if expected == 'pseudo':
                log.error(u'Selector: Pseudo name expected %s.' %
                          selectorText)
                return
            elif expected == 'classname':
                log.error(u'Selector: Class name expected %s.' %
                          selectorText)
                return
            elif newseq and isinstance(newseq[-1], basestring) and \
                 newseq[-1] in COMBINATORS:
                log.error(u'Selector: Cannot end with combinator: %s.' %
                    selectorText)
                return

            if found1:
                self.seq = newseq
            else:
                log.error(
                    u'Selector: No selector found: %s.' % selectorText)
        else:
            log.error(u'Selector: Empty selector.')

    selectorText = property(_getSelectorText, _setSelectorText,
        doc="(DOM) The parsable textual representation of the selector.")


if __name__ == '__main__':
    s = Selector()
    s.selectorText = u'a+ b c'
    print s.selectorText
