#!/usr/bin/env python
"""serializer classes for CSS classes

TODO:
- indent subrules if selector is subselector

"""
__all__ = ['CSSSerializer']
__docformat__ = 'restructuredtext'
__version__ = '0.9a6'

import cssutils

class Preferences(object):
    """
    controls output of CSSSerializer
    """
    def __init__(self, indent=u'    '):
        """
        indent
            indentation of e.g Properties inside a CSSStyleDeclaration
        
        Always use named instead of positional parameters!
        """
        # should the literal @keyword from src CSS be used or the default
        # form, e.g. if True: @im\port else: @import
        self.srcatkeyword = False
        
        # string to use for indentation inside CSSStyleDeclaration and
        # other classes
        self.indent = indent

        # uses hreftype if None or explicit 'string' or 'uri'
        self.importHrefFormat= None

        # only working for complete CSSStyleSheet
        self.linenumbers = False

        # if False removes all CSSComments
        self.keepcomments = True

        # TODO
        self.removeInvalid = True

        # unused yet
        self.setBOM = False
    

class CSSSerializer(object):
    """
    Methods to serialize a CSSStylesheet and its parts

    To use your own serializing method the easiest is to subclass CSS
    Serializer and overwrite the methods you like to customize.
    """
    def __init__(self, prefs=None):
        """
        prefs
            instance of Preferences
        """
        if not prefs:
            prefs = Preferences(u'    ')
            
        self.prefs = prefs
        self.ttypes = cssutils.token.Token
        
        self._map = {
            cssutils.css.cssrule.CSSRule.COMMENT: self.do_CSSComment,
            cssutils.css.cssrule.CSSRule.UNKNOWN_RULE:
                self.do_CSSUnknownRule,
            cssutils.css.cssrule.CSSRule.STYLE_RULE:
                self.do_CSSStyleRule,
            cssutils.css.cssrule.CSSRule.CHARSET_RULE:
                self.do_CSSCharsetRule,
            cssutils.css.cssrule.CSSRule.IMPORT_RULE:
                self.do_CSSImportRule,
            cssutils.css.cssrule.CSSRule.MEDIA_RULE: self.do_CSSMediaRule,
            cssutils.css.cssrule.CSSRule.PAGE_RULE: self._do_unknown
            }
        

    def _serialize(self, text):
        if self.prefs.linenumbers:
            pad = text.count(u'\n') / 10 + 1
            out = []
            for i, line in enumerate(text.split(u'\n')):
                out.append((u'%'+str(pad)+'i: %s') % (i+1, line))
            text = u'\n'.join(out)
        return text


    def _noinvalids(self, x):
        """
        returns
            True if x.valid or prefs.removeInvalid == False
            else False
        """
        if self.prefs.removeInvalid and \
           hasattr(x, 'valid') and not x.valid:
            return True
        else:
            return False


    def _getatkeyword(self, rule, default):
        """
        used by all @rules to get the keyword used
        dependent on prefs setting srcatkeyword
        """
        if self.prefs.srcatkeyword:
            return rule.atkeyword
        else:
            return default

    def _do_unknown(self, rule):
        raise NotImplementedError(
            "Serializer not implemented for %s" % rule)


    def do_stylesheet(self, stylesheet):
        out = []
        for rule in stylesheet.cssRules:
            cssText = self._map.get(rule.type, self._do_unknown)(rule)
            if cssText:
                out.append(cssText)
        return self._serialize(u'\n'.join(out))


    def do_CSSComment(self, rule):
        """
        serializes CSSComment which consists only of commentText
        """
        if self.prefs.keepcomments and rule._cssText:
            return rule._cssText
        else:
            return u''


    def do_CSSUnknownRule(self, rule):
        """
        serializes CSSUnknownRule
        anything until ";" or "{...}"
        + CSSComments
        """
        if rule.atkeyword and not self._noinvalids(rule):
            out = [u'@%s' % rule.atkeyword]
            for part in rule.seq:
                if isinstance(part, cssutils.css.csscomment.CSSComment):
                    if self.prefs.keepcomments:
                        out.append(part.cssText)
                else:
                    out.append(part)
            if not (out[-1].endswith(';') or out[-1].endswith('}')):
                out.append(u';')
            return u''.join(out)
        else:
            return u''


    def do_CSSCharsetRule(self, rule):
        """
        serializes CSSCharsetRule
        encoding: string

        always @charset "encoding";
        no comments or other things allowed!
        """
        if not rule.encoding or self._noinvalids(rule):
            return u''
        return u'@charset "%s";' % rule.encoding
##        out = [u'@charset ']
##        for part in rule.seq:
##            if rule.encoding == part:
##                out.append('"%s"' % part)
##                if isinstance(part, cssutils.css.csscomment.CSSComment):
##                    if self.prefs.keepcomments:
##                 out.append(part.cssText)
##        out.append(';')
##        return u''.join(out)


    def do_CSSImportRule(self, rule):
        """
        serializes CSSImportRule
        
        href 
            string
        hreftype
            'uri' or 'string'
        media
            cssutils.stylesheets.medialist.MediaList
            
        + CSSComments
        """
        if not rule.href or self._noinvalids(rule):
            return u''
        out = [u'%s ' % self._getatkeyword(rule, u'@import')]
        for part in rule.seq:
            if rule.href == part:
                if self.prefs.importHrefFormat == 'string':
                    out.append('"%s"' % part)
                elif self.prefs.importHrefFormat == 'uri':
                    out.append('url(%s)' % part)
                elif rule.hreftype == 'string':
                    out.append('"%s"' % part)
                else:
                    out.append('url(%s)' % part)
            elif isinstance(
                  part, cssutils.stylesheets.medialist.MediaList):
                mediaText = self.do_stylesheets_medialist(part).strip()
                if mediaText and not mediaText == u'all':
                    out.append(u' %s' % mediaText)
            elif hasattr(part, 'cssText'): # comments
                out.append(part.cssText)
        out.append(';')
        return u''.join(out)


    def do_CSSMediaRule(self, rule):
        """
        serializes CSSMediaRule

        + CSSComments
        """
        if not rule.cssRules or self._noinvalids(rule.media):
            return u''        
        mediaText = self.do_stylesheets_medialist(rule.media).strip()
        out = [u'%s %s {\n' % (
            self._getatkeyword(rule, u'@media'), mediaText)]
        for r in rule.cssRules:
            rtext = r.cssText
            if rtext:
                rtext = '\n'.join(
                    [u'%s%s' % (self.prefs.indent, line)
                         for line in rtext.split('\n')])
                out.append(rtext + u'\n')
        out.append('%s}' % self.prefs.indent)
        return u''.join(out)


        indentspace = indent * ' '
        out = []
        out.append(u'@media %s {' % (self._media.getFormatted()))
        for rule in self._rules:
            out.append(rule.getFormatted(indent, indent))
        out.append(u'}')
        return u'\n'.join(out)



    def do_stylesheets_medialist(self, medialist):
        """
        comma-separated list of media, default is 'all'

        If "all" is in the list, every other media *except* "handheld" will
        be stripped. This is because how Opera handles CSS for PDAs.
        """
        if len(medialist.seq) == 0:
            return u'all'
        else:
            #import pprint;pprint.pprint( medialist.seq)
            hasall = bool(
                [1 for part in medialist.seq if part == u'all'])
            hashandheld = bool(
                [1 for part in medialist.seq if part == u'handheld'])
            doneall = donehandheld = False
            out = []
            for part in medialist.seq:
                if hasattr(part, 'cssText'): # comments
                    out.append(part.cssText)
                elif u',' == part and not hasall:
                    out.append(u', ')
                elif not hasall or (
                           hasall and hashandheld and \
                           part in (u'handheld', u'all')):
                    if hasall and hashandheld and (
                        (donehandheld and part == u'all') or\
                        (doneall and part == u'handheld')
                         ):
                        out.append(u', ')
                    if part == u'all':
                        doneall = True
                    else:
                        donehandheld = True
                    out.append(part)
            if out == []:
                out = [u'all']
            return u''.join(out)


    def do_CSSStyleRule(self, rule):
        """
        serializes CSSCharsetRule

        encoding
            string
        
        + CSSComments
        """
        selectorText = self.do_css_SelectorList(rule.selectorList)
        if not selectorText:
            return u''
        styleText = self.do_css_CSSStyleDeclaration(rule.style)
        if not styleText:
            styleText = u''
        
        return u'%s {%s}' % (selectorText, styleText)


    def do_css_SelectorList(self, selectorlist):
        """
        comma-separated list of Selectors
        """
        if len(selectorlist.seq) == 0 or self._noinvalids(selectorlist):
            return u''
        else:
            out = []
            for part in selectorlist.seq:
                if hasattr(part, 'cssText'): # comments
                    out.append(part.cssText)
                elif u',' == part:
                    out.append(u', ')
                elif isinstance(part, cssutils.css.cssstylerule.Selector):
                    out.append(self.do_css_Selector(part).strip())
                else:
                    out.append(part) # ?
            return u''.join(out)


    def do_css_Selector(self, selector):
        """
        a single selector including comments

        TODO: style combinators like + >        
        """
        if len(selector.seq) == 0 or self._noinvalids(selector):
            return u''
        else:
            out = []
            for part in selector.seq:
                if hasattr(part, 'cssText'): # comments
                    out.append(part.cssText)
                else:
                    out.append(part)
            return u''.join(out)


    def do_css_CSSStyleDeclaration(self, style):
        """
        Style declaration of CSSStyleRule
        """
        if len(style.seq) == 0 or self._noinvalids(style):
            return u''
        else:
            out = []
            done1 = False
            for part in style.seq:
                if isinstance(part, cssutils.css.csscomment.CSSComment):
                    if self.prefs.keepcomments:
                        out.append(self.prefs.indent)
                        out.append(part.cssText)
                        out.append(u'\n')
                elif isinstance(
                        part, cssutils.css.cssstyledeclaration.Property):
                    propertytext = self.do_css_Property(part)
                    if propertytext:
                        if done1:
                            out.append(u';\n')
                        done1 = True
                        out.append(self.prefs.indent)
                        out.append(propertytext)
                else:
                    out.append(part)
            if done1:
                out.insert(0, u'\n')
                out.append(u'\n')
                out.append(self.prefs.indent)
            return u''.join(out)


    def do_css_Property(self, property):
        """
        Style declaration of CSSStyleRule

        Property has a seqs attribute which contains 3 seq lists for
        name, value and priority
        """
        if not property.seqs[0] or self._noinvalids(property):
            return u''
        else:
            out = []
            nameseq, valueseq, priorityseq = property.seqs

            for part in nameseq:
                if hasattr(part, 'cssText'): # comments
                    out.append(part.cssText)
                else:
                    out.append(part)
            if out:
                out.append(u': ')
            v = self.do_css_Propertyvalue(valueseq)
            if v:
                out.append(v)

            if out and priorityseq:
                out.append(u' ')
            for part in priorityseq:
                if hasattr(part, 'cssText'): # comments
                    out.append(part.cssText)
                else:
                    out.append(part)
            
            return u''.join(out)


    def do_css_Propertyvalue(self, valueseq):
        """
        serializes a property
        """
        if not valueseq:
            return u''
        else:
            out = []
            for part in valueseq:
                if hasattr(part, 'cssText'): # comments
                    out.append(part.cssText)
                else:
                    out.append(part)
            return u''.join(out).strip()
