"""
tests for css.CSSStyleSheet
"""
__version__ = '0.9a6'

import xml.dom

import basetest

import cssutils.css

class CSSStyleSheetTestCase(basetest.BaseTestCase):

    def setUp(self):
        super(CSSStyleSheetTestCase, self).setUp()

        self.r = cssutils.css.CSSStyleSheet() # used by basetest
        self.s = self.r # used here

        self.rule = cssutils.css.CSSStyleRule()        


    def test_init(self):
        "CSSStyleSheet.__init__()"
        self.assertEqual(False, self.s._readonly)
        self.assertEqual([], self.s.cssRules)
        self.assertEqual('text/css', self.s.type)

        self.assertEqual(False, self.s.disabled)
        self.assertEqual(None, self.s.href)
        self.assertEqual(None, self.s.media)
        self.assertEqual(None, self.s.ownerNode)
        self.assertEqual(None, self.s.parentStyleSheet)
        self.assertEqual(u'', self.s.title)

    def test_NoModificationAllowedErr(self):
        "CSSStyleSheet NoModificationAllowedErr"
        css = cssutils.css.CSSStyleSheet(readonly=True)
        
        self.assertEqual(True, css._readonly) # internal...
        
        self.assertRaises(xml.dom.NoModificationAllowedErr,
                          css._setCssText, u'@x;')
        self.assertRaises(xml.dom.NoModificationAllowedErr,
                          css.insertRule, self.rule)
        self.assertRaises(xml.dom.NoModificationAllowedErr,
                          css.insertRule, self.rule, 0)
        self.assertRaises(xml.dom.NoModificationAllowedErr,
                          css.deleteRule, 0)


    def test_cssText_HierarchyRequestErr(self):
        "CSSStyleSheet.cssText HierarchyRequestErr"
        tests = {
            # @charset: only one and 1st
            u' @charset "utf-8";': xml.dom.HierarchyRequestErr,
            u'/*c*/@charset "ascii";': xml.dom.HierarchyRequestErr,
            u'@charset "ascii";@charset "ascii";': xml.dom.HierarchyRequestErr,
            u'@import "x"; @charset "ascii";': xml.dom.HierarchyRequestErr,
            u'@media all {} @charset "ascii";': xml.dom.HierarchyRequestErr,
            u'a {} @charset "ascii";': xml.dom.HierarchyRequestErr,
            # @import: before @media, stylerule
            u'@media all {} @import "x";': xml.dom.HierarchyRequestErr,
            u'a {} @import "x";': xml.dom.HierarchyRequestErr,
            }
        self.do_raise_r(tests)
        self.do_raise_p(tests)

    def test_cssText_SyntaxErr(self):
        """CSSStyleSheet.cssText SyntaxErr

        for single {, } or ;
        """
        tests = {
            u'{': xml.dom.SyntaxErr,
            u'}': xml.dom.SyntaxErr,
            u';': xml.dom.SyntaxErr,
            u'@charset "ascii";{': xml.dom.SyntaxErr,
            u'@charset "ascii";}': xml.dom.SyntaxErr,
            u'@charset "ascii";;': xml.dom.SyntaxErr,
            }
        self.do_raise_r(tests)
        self.do_raise_p(tests)


    def test_cssText(self):
        "CSSStyleSheet.cssText"
        tests = {
            u'': None,
            # @charset
            u'@charset "ascii";\n@import "x";': None,
            u'@charset "ascii";\n@media all {}': u'@charset "ascii";',
            u'@charset "ascii";\n@x;': None,
            u'@charset "ascii";\na {}': None,
            # @import
            u'@x;\n@import "x";': None,
            u'@import "x";\n@import "y";': None,
            u'@import "x";\n@media all {}': u'@import "x";', 
            u'@import "x";\n@x;': None,
            u'@import "x";\na {}': None,
            }
        self.do_equal_r(tests)
        self.do_equal_p(tests)


    def test_deleteRule(self):
        "CSSStyleSheet.deleteRule()"
        self.s.cssText = u'@charset "ascii"; @import "x"; @x; a {}@y;'
        self.assertEqual(5, self.s.cssRules.length)
        
        self.assertRaises(xml.dom.IndexSizeErr, self.s.deleteRule, 5)

        # end -1
        self.s.deleteRule(-1)
        self.assertEqual(4, self.s.cssRules.length)
        self.assertEqual(
            u'@charset "ascii";\n@import "x";\n@x;\na {}', self.s.cssText)
        # beginning
        self.s.deleteRule(0)
        self.assertEqual(3, self.s.cssRules.length)
        self.assertEqual(u'@import "x";\n@x;\na {}', self.s.cssText)
        # middle
        self.s.deleteRule(1)
        self.assertEqual(2, self.s.cssRules.length)
        self.assertEqual(u'@import "x";\na {}', self.s.cssText)
        # end
        self.s.deleteRule(1)
        self.assertEqual(1, self.s.cssRules.length)
        self.assertEqual(u'@import "x";', self.s.cssText)
        

    def test_insertRule(self):
        "CSSStyleSheet.insertRule()"
        # prepare
        s = cssutils.css.CSSStyleSheet()
        charsetrule = cssutils.css.CSSCharsetRule('ascii')
        importrule = cssutils.css.CSSImportRule('x')
        mediarule = cssutils.css.CSSMediaRule()
        unknownrule = cssutils.css.CSSUnknownRule('@x;')
        stylerule = cssutils.css.CSSStyleRule('a')
        comment = cssutils.css.CSSComment(u'/*c*/')

        s.insertRule(stylerule, 0)
        s.insertRule(unknownrule, 0)
        self.assertEqual(2, len(s.cssRules))
        self.assertEqual(
                u'@x;\na {}', s.cssText)        

        # HIERARCHY_REQUEST_ERR
        # @import: 
        self.assertRaises(xml.dom.HierarchyRequestErr,
                          s.insertRule, importrule)
        self.assertEqual(2, len(s.cssRules))
        # after @x
        s.insertRule(importrule, 1)
        # before @x
        s.insertRule(importrule, 0)
        self.assertEqual(u'@import url(x);\n@x;\n@import url(x);\na {}',
                         s.cssText)

        # @charset new
        s = cssutils.css.CSSStyleSheet()
        s.insertRule(importrule)
        s.insertRule(unknownrule)
        s.insertRule(stylerule)
        self.assertEqual(
                u'@import url(x);\n@x;\na {}', s.cssText)
        
        # @charset: only at beginning
        self.assertRaises(xml.dom.HierarchyRequestErr,
                          s.insertRule, charsetrule, 1)
        self.assertEqual(3, len(s.cssRules))
        # @charset insert 1st
        s.insertRule(charsetrule, 0)
        self.assertEqual(4, len(s.cssRules))
        # @charset: no 2nd
        self.assertRaises(xml.dom.HierarchyRequestErr,
                          s.insertRule, charsetrule, 0)
        self.assertEqual(4, len(s.cssRules))
        self.assertEqual(u'@charset "ascii";\n@import url(x);\n@x;\na {}',
                         s.cssText)
        
        # nothing before @charset
        self.assertRaises(xml.dom.HierarchyRequestErr,
                          s.insertRule, charsetrule, 0)
        self.assertRaises(xml.dom.HierarchyRequestErr,
                          s.insertRule, importrule, 0)
        self.assertRaises(xml.dom.HierarchyRequestErr,
                          s.insertRule, mediarule, 0)
        self.assertRaises(xml.dom.HierarchyRequestErr,
                          s.insertRule, stylerule, 0)
        self.assertRaises(xml.dom.HierarchyRequestErr,
                          s.insertRule, comment, 0)

        self.assertRaises(xml.dom.HierarchyRequestErr,
                          s.insertRule, charsetrule, 1)

        # index
        self.assertRaises(xml.dom.IndexSizeErr,
                  s.insertRule, stylerule, -1)
        self.assertRaises(xml.dom.IndexSizeErr,
                  s.insertRule, stylerule, s.cssRules.length + 1)


if __name__ == '__main__':
    import unittest
    unittest.main()
